import os

import numpy as np

from PyQt5 import QtCore

from vision.utils import polygon_center, center_for_multipin
from .pin import Pin


__all__ = ["Element"]


class Element:
    def __init__(self, pins=None, name="", manual_name="", bounding_zone=None, probability=None,
                 rotation=0, w_pins=-1, h_pins=-1, width=-1, height=-1,
                 is_manual=False, center=None, side_idxes=None):
        self.pins = pins or []
        self.name = name
        self.manual_name = manual_name
        self.bounding_zone = np.array(bounding_zone if bounding_zone is not None else []).astype(np.float64)
        self.probability = probability
        self.rotation = rotation
        self.w_pins = w_pins
        self.h_pins = h_pins
        self.width = width
        self.height = height
        self.is_manual = is_manual
        self.side_idxes = side_idxes
        if center is None:
            if w_pins > 0 or h_pins > 0:
                self.center = center_for_multipin(pins)
            else:
                self.center = polygon_center(self.bounding_zone)
        else:
            self.center = center
        self.getSizes()
        self.adjust_size_for_pins()

    def to_json(self) -> dict:
        side_indexes = \
            tuple(map(lambda p: tuple(map(int, p)), self.side_idxes)) if self.side_idxes is not None else None
        return {
            "pins": [pin.to_json() for pin in self.pins],
            "name": str(self.name),
            "manual_name": str(self.manual_name),
            "bounding_zone": tuple(tuple(map(float, p)) for p in self.bounding_zone),
            "probability": float(self.probability) if self.probability is not None else None,
            "rotation": int(self.rotation),
            "w_pins": int(self.w_pins),
            "h_pins": int(self.h_pins),
            "width": float(self.width),
            "height": float(self.height),
            "is_manual": bool(self.is_manual),
            "side_indexes": side_indexes,
            "center": (float(self.center[0]), float(self.center[1]))
        }

    @classmethod
    def from_json(cls, d: dict) -> "Element":
        return cls(
            pins=list(map(Pin.from_json, d["pins"])),
            name=d["name"],
            manual_name=d["manual_name"],
            bounding_zone=np.array(d["bounding_zone"]),
            probability=float(d["probability"]) if d["probability"] is not None else None,
            rotation=d["rotation"],
            w_pins=d["w_pins"],
            h_pins=d["h_pins"],
            width=d["width"],
            height=d["height"],
            is_manual=d["is_manual"],
            side_idxes=d["side_indexes"],
            center=np.array(d["center"])
        )

    def clear_ivc(self):
        for pin in self.pins:
            del pin.ivc

    def clear_reference_ivc(self):
        for pin in self.pins:
            del pin.reference_ivc

    def getSizes(self):
        if self.is_manual:
            return
        pix_per_mm = 23

        # default border - rectangle around pins
        min_x = min([pin.c for pin in self.pins]) if len(self.pins) else 0
        max_x = max([pin.c for pin in self.pins]) if len(self.pins) else 0
        min_y = min([pin.r for pin in self.pins]) if len(self.pins) else 0
        max_y = max([pin.r for pin in self.pins]) if len(self.pins) else 0

        self.width = max_x - min_x + pix_per_mm * 2
        self.height = max_y - min_y + pix_per_mm * 2
        if self.rotation % 2 == 0:
            self.width, self.height = self.height, self.width

        with open(os.path.join("svg", "sizes.csv")) as file:
            for line in file.readlines():
                if line.split(",")[0].strip() == self.reprName():
                    self.width = float(line.split(",")[1].strip()) * pix_per_mm
                    self.height = float(line.split(",")[2].strip()) * pix_per_mm
                    break

    def adjust_size_for_pins(self):
        """Adjusts element rect to cover all pins."""

        if len(self.pins) > 0:
            pin_coordinates = np.array(tuple(pin.rc for pin in self.pins))
            pin_distances = np.abs(pin_coordinates - np.asarray(self.center))
            max_distance = np.max(pin_distances, axis=0)

            size = (self.width, self.height) if self.rotation % 2 == 0 else (self.height, self.width)
            scale_factor = np.max(2 * max_distance / np.asarray(size))

            if scale_factor > 1.0:  # Don't decrease element size
                self.width *= scale_factor
                self.height *= scale_factor

    def getRect(self):
        return QtCore.QRect(self.center[1] - self.width / 2, self.center[0] - self.height / 2, self.width, self.height)

    def reprName(self):
        name = self.name.split("/")[0].split("\\")[0]
        if self.w_pins >= 0:
            return name % (self.w_pins + self.h_pins)
        else:
            return name

    def __repr__(self):
        if len(self.pins) <= 2:
            return self.reprName() + ": " + " ".join(["(%.2f, %.2f)" % (pin.r, pin.c) for pin in self.pins])
        else:
            result = self.reprName() + ":\n"
            for i in range(0, len(self.pins), 2):
                result += "  " + " ".join(["(%.2f, %.2f)" % (pin.r, pin.c) for pin in self.pins[i: i + 2]])
                if i + 2 < len(self.pins):
                    result += "\n"
            return result

    def is_rotatable(self):
        return self.side_idxes is not None

    def rotate_pins(self):
        if self.side_idxes is None:
            return
        assert (self.side_idxes[0][0] == 0)
        self.pins = self.pins[self.side_idxes[0][1]:] + self.pins[:self.side_idxes[0][1]]
        self.side_idxes = self.side_idxes[1:] + self.side_idxes[:1]
        i = 0
        for k in range(len(self.side_idxes)):
            side_len = self.side_idxes[k][1] - self.side_idxes[k][0]
            self.side_idxes[k] = (i, i + side_len)
            i += side_len

    def removePin(self, idx):
        if self.side_idxes is None:
            self.pins.pop(idx)
            return
        i = 0
        while self.side_idxes[i][1] <= idx:
            i += 1
        self.side_idxes[i] = self.side_idxes[i][0], self.side_idxes[i][1] - 1
        if self.side_idxes[i][0] == self.side_idxes[i][1]:
            self.side_idxes.pop(i)
        else:
            i += 1
        while i < len(self.side_idxes):
            self.side_idxes[i] = self.side_idxes[i][0] - 1, self.side_idxes[i][1] - 1
            i += 1
        self.pins.pop(idx)
